# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors.compressors import ModelCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationArgs, QuantizationScheme
from compressed_tensors.quantization.lifecycle.initialize import (
    initialize_module_for_quantization,
)
from compressed_tensors.quantization.quant_args import FP8_E4M3_DATA
from tests.testing_utils import requires_gpu
from transformers import AutoModelForCausalLM, AutoTokenizer


@pytest.mark.parametrize(
    "frozen_stub,q_format,compressed_stub",
    [
        (
            "nm-testing/llama2.c-stories42M-gsm8k-quantized-only-uncompressed",
            "float-quantized",
            "nm-testing/llama2.c-stories42M-gsm8k-quantized-only-compressed",
        ),
        (
            "nm-testing/llama2.c-stories15M-ultrachat-mixed-uncompressed",
            "pack-quantized",
            "nm-testing/llama2.c-stories15M-ultrachat-mixed-compressed",
        ),
    ],
)
def test_compress_model(frozen_stub, q_format, compressed_stub):
    """Check that compression generates the expected compressed model"""
    model = AutoModelForCausalLM.from_pretrained(frozen_stub, torch_dtype=torch.float32)
    compressor = ModelCompressor.from_pretrained_model(model, None, q_format)
    true_compressed_model = AutoModelForCausalLM.from_pretrained(
        compressed_stub, torch_dtype=torch.float32
    )

    compressor.compress_model(model)
    compressed = dict(model.state_dict())
    true_compressed = dict(true_compressed_model.state_dict())

    assert compressed.keys() == true_compressed.keys()
    for key in compressed.keys():
        assert compressed[key].dtype == true_compressed[key].dtype, key
        assert torch.equal(compressed[key], true_compressed[key])


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    "model_stub,q_format,compressed_stub",
    [
        (
            "nm-testing/llama2.c-stories42M-gsm8k-quantized-only-uncompressed",
            "float-quantized",
            "nm-testing/llama2.c-stories42M-gsm8k-quantized-only-compressed",
        ),
        (
            "nm-testing/llama2.c-stories15M-ultrachat-mixed-uncompressed",
            "pack-quantized",
            "nm-testing/llama2.c-stories15M-ultrachat-mixed-compressed",
        ),
    ],
)
def test_decompress_model(model_stub, q_format, compressed_stub):
    from transformers.utils.quantization_config import CompressedTensorsConfig

    model = AutoModelForCausalLM.from_pretrained(model_stub, torch_dtype=torch.float32)
    compressor = ModelCompressor.from_pretrained_model(model, None, q_format)
    true_decompressed_model = AutoModelForCausalLM.from_pretrained(
        compressed_stub,
        quantization_config=CompressedTensorsConfig(run_compressed=False),
        torch_dtype=torch.float32,
    )

    compressor.compress_model(model)
    compressor.decompress_model(model)

    true_decompressed = dict(true_decompressed_model.state_dict())
    decompressed = dict(model.state_dict())

    assert decompressed.keys() == true_decompressed.keys()
    for key in decompressed.keys():
        assert decompressed[key].dtype == true_decompressed[key].dtype, key
        assert torch.equal(decompressed[key], true_decompressed[key])


def test_multiple_quant_compressors():
    model = torch.nn.Sequential(torch.nn.Linear(1, 2), torch.nn.Linear(2, 3))
    input_activations = QuantizationArgs(num_bits=8, type="float")
    weights = QuantizationArgs(num_bits=8, type="float")

    scheme_fp8 = QuantizationScheme(
        targets=["Linear"],
        weights=weights,
        input_activations=input_activations,
        format=CompressionFormat.float_quantized.value,
    )

    input_activations = QuantizationArgs(
        num_bits=4,
        type="float",
        scale_dtype=FP8_E4M3_DATA.dtype,
        zp_dtype=FP8_E4M3_DATA.dtype,
    )
    weights = QuantizationArgs(
        num_bits=4,
        type="float",
        scale_dtype=FP8_E4M3_DATA.dtype,
        zp_dtype=FP8_E4M3_DATA.dtype,
    )

    scheme_nvfp4 = QuantizationScheme(
        targets=["Linear"],
        weights=weights,
        input_activations=input_activations,
        format=CompressionFormat.nvfp4_pack_quantized.value,
    )

    model[0].quantization_scheme = scheme_fp8
    initialize_module_for_quantization(model[0])
    model[0].quantization_status = "frozen"
    model[1].quantization_scheme = scheme_nvfp4
    initialize_module_for_quantization(model[1])
    model[1].quantization_status = "frozen"

    compressor = ModelCompressor.from_pretrained_model(model, None)
    compressor.compress_model(model)
    assert compressor.quantization_config.format == CompressionFormat.mixed_precision
    assert model[0].quantization_scheme.format == scheme_fp8.format
    assert model[1].quantization_scheme.format == scheme_nvfp4.format


@requires_gpu
def test_compressed_model_inference_with_hook():
    model_stub = "nm-testing/llama2.c-stories42M-gsm8k-quantized-only-compressed"

    # Load compressed model
    model = AutoModelForCausalLM.from_pretrained(
        model_stub, dtype="auto", device_map="auto"
    )
    tokenizer = AutoTokenizer.from_pretrained(model_stub)

    # Model should have the decompression hook attached
    assert hasattr(model, "ct_decompress_hook")

    # Run a forward pass to trigger the hook
    prompt = "The quick brown fox jumps over the lazy dog"
    inputs = tokenizer(prompt, return_tensors="pt")
    input_ids = inputs["input_ids"].to(device=model.device)

    with torch.no_grad():
        outputs = model(input_ids=input_ids, labels=input_ids)

    # After forward pass, hook should have triggered and been removed
    assert not hasattr(model, "ct_decompress_hook")

    # Check perplexity is reasonable
    assert torch.exp(outputs.loss) <= 500.0



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import Callable, TypeVar


T = TypeVar("T")


__all__ = ["SearchFailureError", "max_binary_search"]


class SearchFailureError(ValueError):
    pass


def max_binary_search(
    fn: Callable[[int], T],
    cond: Callable[[T], bool],
    start: int,
    end: int,
) -> tuple[int, T]:
    best_idx = None
    best_val = None

    while start <= end:
        mid = (start + end) // 2
        val = fn(mid)

        if cond(val):
            # condition is true, search higher
            best_idx, best_val = mid, val
            start = mid + 1
        else:
            # condition is false, search lower
            end = mid - 1

    if best_idx is None:
        raise SearchFailureError()

    return best_idx, best_val



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.transform import HadamardFactory, TransformFactory
from compressed_tensors.transform.utils.hadamard import random_hadamard_matrix
from torch import device, dtype
from torch.nn import Parameter


@TransformFactory.register("random-hadamard")
class RandomHadamardFactory(HadamardFactory):
    """
    Factory used to apply random hadamard transforms to a model

    :param name: name associated with transform scheme
    :param scheme: transform scheme which defines how transforms should be created
    :param seed: random seed used to transform weight randomization
    """

    def _create_weight(
        self,
        size: int,
        device: device,
        construct_device: device,
        precision: dtype,
    ) -> Parameter:
        data = random_hadamard_matrix(size, precision, construct_device, self.generator)
        data = data.to(device=device)
        return Parameter(data, requires_grad=self.scheme.requires_grad)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
# isort: off
from .logger import LoggerConfig, configure_logger, logger
from .base import *

from .compressors import *
from .config import *
from .quantization import QuantizationConfig, QuantizationStatus

# avoid resolving compressed_tensors.offload as compressed_tensors.utils.offload
from .utils.offload import *
from .utils.helpers import *
from .utils.internal import *
from .utils.match import *
from .utils.permutations_24 import *
from .utils.safetensors_load import *
from .utils.semi_structured_conversions import *
from .utils.type import *
from .version import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.transform import TransformScheme
from pydantic import BaseModel, ConfigDict


__all__ = ["TransformConfig"]


class TransformConfig(BaseModel):
    """
    Configuration of transforms to be applied to a model. This config is to be
    serialized within a model's `config.json` file

    :param config_groups: A dictionary of `TransformSchemes` that should be applied
        to a particular model. The keys can be any arbitrary string
    """

    config_groups: dict[str, TransformScheme]

    model_config = ConfigDict(extra="forbid")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import logging
import math

import torch
from compressed_tensors.quantization.quant_args import (
    FP4_E2M1_DATA,
    FP8_E4M3_DATA,
    FloatArgs,
    QuantizationArgs,
    QuantizationStrategy,
    QuantizationType,
    round_to_quantized_type_dtype,
)
from compressed_tensors.quantization.utils.mxfp_utils import (
    generate_mx_scales,
    maybe_convert_from_mx_exp,
    should_generate_mx_scales,
)
from loguru import logger
from torch import FloatTensor, IntTensor, Tensor
from torch.nn import Module


__all__ = [
    "is_module_quantized",
    "is_model_quantized",
    "module_type",
    "get_torch_bit_depth",
    "can_quantize",
    "KV_CACHE_TARGETS",
    "compute_dynamic_scales_and_zp",
    "calculate_range",
    "calculate_qparams",
    "generate_gparam",
    "strategy_cdiv",
    "calculate_block_padding",
    "maybe_pad_tensor_for_block_quant",
]

# target the self_attn layer
# QuantizedKVParameterCache is responsible for obtaining the k_scale and v_scale
KV_CACHE_TARGETS = ["re:.*self_attn$"]

_LOGGER: logging.Logger = logging.getLogger(__name__)


def calculate_qparams(
    min_vals: Tensor,
    max_vals: Tensor,
    quantization_args: QuantizationArgs,
    global_scale: Tensor | None = None,
) -> tuple[FloatTensor, IntTensor]:
    """
    :param min_vals: tensor of min value(s) to calculate scale(s) and zero point(s)
        from
    :param max_vals: tensor of max value(s) to calculate scale(s) and zero point(s)
        from
    :param quantization_args: settings to quantization
    :param global_scale: additional global scale to scale the locally generated scale
        currently only applied/supported for Fp4

    :return: tuple of the calculated scale(s) and zero point(s). For FP4, the calculated
        scale is of dtype FP8
    """
    # based on the implementations for consuming quantized values,
    # 0.0 must always be representable within the quantized range
    min_vals = torch.min(min_vals, torch.zeros_like(min_vals))
    max_vals = torch.max(max_vals, torch.zeros_like(max_vals))

    device = min_vals.device

    bit_min, bit_max = calculate_range(quantization_args, device)
    bit_range = bit_max - bit_min

    # 1. Generate scale and zero-point
    if quantization_args.symmetric:
        max_val_pos = torch.max(torch.abs(min_vals), torch.abs(max_vals))
        if should_generate_mx_scales(args=quantization_args):
            scales = generate_mx_scales(
                x=max_val_pos, num_bits=quantization_args.num_bits
            )
        else:
            scales = max_val_pos / (float(bit_range) / 2)
        zero_points = torch.zeros(scales.shape, device=device, dtype=min_vals.dtype)
    else:
        if (
            quantization_args.num_bits == 4
            and quantization_args.type == QuantizationType.FLOAT
        ):
            raise NotImplementedError(
                "Asymmetric Quantization is not supported for FP4"
            )
        scales = (max_vals - min_vals) / float(bit_range)
        zero_points = bit_min - (min_vals / scales)
        zero_points = torch.clamp(zero_points, bit_min, bit_max)

    # 2. Conditionally scale the generated local scale by a global_scale
    if global_scale is not None:
        scales = global_scale * scales

    # 3. Conditionally round the scale to the quantized dtype, if scale_dtype is set
    if quantization_args.scale_dtype is not None:
        scales = round_to_quantized_type_dtype(
            scales, dtype=quantization_args.scale_dtype
        )

    # 4. Optionally remove exponent
    scales = maybe_convert_from_mx_exp(quantization_args, scales)

    # 5. Update any 0s with small values to
    # prevent div by 0
    eps = _get_dtype_eps(
        dtype=(
            quantization_args.scale_dtype
            if quantization_args.scale_dtype is not None
            else scales.dtype
        )
    )
    scales = torch.where(
        scales == 0,
        torch.tensor(eps, dtype=scales.dtype, device=device),
        scales,
    )

    # 6. Round the zp to zp_dtype
    zero_points = round_to_quantized_type_dtype(
        zero_points, dtype=quantization_args.zp_dtype, cast_to_original_dtype=False
    )

    if scales.ndim == 0:
        scales = scales.reshape(1)
        zero_points = zero_points.reshape(1)

    return scales, zero_points


def compute_dynamic_scales_and_zp(
    value: Tensor,
    args: QuantizationArgs,
    module: torch.nn.Module,
    global_scale: Tensor | None = None,
):
    """
    Returns the computed scales and zero points for dynamic activation
    quantization.

    :param value: tensor to calculate quantization parameters for
    :param args: quantization args
    :param reduce_dims: optional tuple of dimensions to reduce along,
        returned scale and zero point will be shaped (1,) along the
        reduced dimensions
    :return: tuple of scale and zero point derived from the observed tensor
    """

    keep_dims = True
    if args.strategy == QuantizationStrategy.TOKEN:
        dim = {0, 1}
        reduce_dims = tuple(idx for idx in range(value.ndim) if idx not in dim)
    elif args.strategy == QuantizationStrategy.TENSOR:
        reduce_dims = None
    elif args.strategy in (
        QuantizationStrategy.TENSOR_GROUP,
        QuantizationStrategy.GROUP,
    ):

        reduce_dims = -1
        keep_dims = False

        reshaped_dims = (
            math.ceil(value.shape[-1] / args.group_size),
            args.group_size,
        )
        value = value.unflatten(-1, reshaped_dims)

    else:
        supported_strategies = (
            QuantizationStrategy.TOKEN,
            QuantizationStrategy.TENSOR,
            QuantizationStrategy.TENSOR_GROUP,
            QuantizationStrategy.GROUP,
        )
        raise ValueError(
            "Dynamic quantization is only supported for ",
            f"{supported_strategies}",
        )

    if not reduce_dims:
        min_val, max_val = torch.aminmax(value)
    else:
        min_val = torch.amin(value, dim=reduce_dims, keepdims=keep_dims)
        max_val = torch.amax(value, dim=reduce_dims, keepdims=keep_dims)

    return calculate_qparams(min_val, max_val, args, global_scale=global_scale)


def calculate_range(
    quantization_args: QuantizationArgs, device: str
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Calculated the effective quantization range for the given Quantization Args

    :param quantization_args: quantization args to get range of
    :param device: device to store the range to
    :return: tuple endpoints for the given quantization range
    """
    if quantization_args.type == QuantizationType.INT:
        bit_range = 2.0**quantization_args.num_bits
        q_max = torch.tensor(bit_range / 2 - 1, device=device)
        q_min = torch.tensor(-bit_range / 2, device=device)
    elif quantization_args.type == QuantizationType.FLOAT:
        if quantization_args.num_bits == 8:
            q_max = torch.tensor(FP8_E4M3_DATA.max, device=device)
            q_min = torch.tensor(FP8_E4M3_DATA.min, device=device)
        elif quantization_args.num_bits == 4:
            q_max = torch.tensor(FP4_E2M1_DATA.max, device=device)
            q_min = torch.tensor(FP4_E2M1_DATA.min, device=device)
        else:
            raise NotImplementedError(
                "Range calculation only supported for 4 and 8 bits"
            )
    else:
        raise ValueError(f"Invalid quantization type {quantization_args.type}")

    return q_min, q_max


def is_module_quantized(module: Module) -> bool:
    """
    Check if a module is quantized, based on the existence of a non-empty quantization
    scheme

    :param module: pytorch module to check
    :return: True if module is quantized, False otherwise
    """
    if not hasattr(module, "quantization_scheme"):
        return False

    if module.quantization_scheme.weights is not None:
        return True

    if module.quantization_scheme.input_activations is not None:
        return True

    if module.quantization_scheme.output_activations is not None:
        return True

    return False


def is_model_quantized(model: Module) -> bool:
    """
    Check if any modules in a model are quantized, based on the existence of a non-empty
    quantization scheme in at least one module

    :param model: pytorch model
    :return: True if model is quantized, False otherwise
    """
    return any(is_module_quantized(submodule) for submodule in model.modules())


def module_type(module: Module) -> str:
    """
    Gets a string representation of a module type

    :module: pytorch module to get type of
    :return: module type as a string
    """
    return type(module).__name__


def get_torch_bit_depth(value: torch.Tensor) -> int:
    """
    Determine the number of bits used to represent the dtype of a tensor

    :param value: tensor to check bit depth of
    :return: bit depth of each element in the value tensor
    """
    try:
        bit_depth = torch.finfo(value.dtype).bits
    except TypeError:
        bit_depth = torch.iinfo(value.dtype).bits

    return bit_depth


def can_quantize(value: torch.Tensor, quant_args: "QuantizationArgs") -> bool:  # noqa
    """
    Checks if value can be quantized by quant_args.

    :param value: tensor to check for quantization
    :param quant_args: QuantizationArgs to use for quantization
    :return: False if value is already quantized to quant_args or value is incompatible
    with quant_args, True if value can be quantized with quant_args
    """
    bit_depth = get_torch_bit_depth(value)
    requested_depth = quant_args.num_bits
    if bit_depth < quant_args.num_bits:
        _LOGGER.warn(
            f"Can't quantize tensor with bit depth {bit_depth} to {requested_depth}."
            "The QuantizationArgs provided are not compatible with the input tensor."
        )

    return bit_depth > quant_args.num_bits


def generate_gparam(
    updated_min_val: torch.Tensor,
    updated_max_val: torch.Tensor,
    scale_data: FloatArgs | None = FP8_E4M3_DATA,
    quant_data: FloatArgs | None = FP4_E2M1_DATA,
    dtype: torch.dtype | None = torch.float32,
):
    """
    Generate a global scale for an entire tensor (input_tensor).
    Goal of the scale is to ensure that the quantization (local) scale
    falls into the approproiate dtype range.

    E.g. for NVFP4, group (local) scales are in dtype FP8. The global_scale
    attempts to use the entire FP8 dtype range while mapping a per-group max
    to the FP4 max.
    """
    min_vals = torch.min(updated_min_val, torch.zeros_like(updated_min_val))
    max_vals = torch.max(updated_max_val, torch.zeros_like(updated_max_val))
    max_val_pos = torch.max(torch.abs(min_vals), torch.abs(max_vals))
    max_val_pos = torch.clamp(max_val_pos, min=torch.finfo(max_val_pos.dtype).tiny)
    global_scale = scale_data.max * quant_data.max / max_val_pos

    # Replace any NaN or Inf with 1.0. NaN arises when max_val_pos was NaN
    # (clamp does not propagate NaN, so it passes through to the division).
    # Inf arises when max_val_pos was near zero and the division overflows float32.
    # global_scale=1 means no global scaling, which is a safe fallback for
    # uncalibrated experts.
    global_scale = torch.nan_to_num(global_scale, nan=1.0, posinf=1.0, neginf=1.0)

    return global_scale.to(dtype).reshape([1])


def strategy_cdiv(
    value: int,
    divisor: int,
    strategy: QuantizationStrategy | None,
    strict: bool = False,
) -> int:
    dividend = math.ceil(value / divisor)
    if dividend * divisor != value:
        message = (
            f"{strategy} quantization strategy requires strict division of "
            f"weight/activation size {value} and group/block size {divisor}. "
            "consider reducing the group/block size or ignoring modules with "
            f"weights not divisible by {divisor}"
        )
        if strict:
            raise ValueError(message)

        else:
            logger.bind(log_once=True).warning(message)

    return dividend


def _get_dtype_eps(dtype: torch.dtype) -> float:
    if dtype == FP8_E4M3_DATA.dtype:
        return 0.125
    elif dtype == FP4_E2M1_DATA.dtype:
        return 0.25
    elif torch.is_floating_point(torch.tensor([], dtype=dtype)):
        return torch.finfo(dtype).eps
    else:
        return 1


def calculate_block_padding(
    shape: tuple[int, ...],
    block_structure: tuple[int, int],
) -> tuple[int, int]:
    """
    Calculate the padding needed to make tensor dimensions divisible by block size.

    For block quantization, dimensions must be divisible by the block size for
    proper scale alignment when layers are merged in inference frameworks like vLLM.

    :param shape: shape of the tensor (at least 2D)
    :param block_structure: [block_height, block_width] for block quantization
    :return: tuple of (pad_rows, pad_cols) needed to make dimensions divisible
    """
    if len(shape) < 2:
        raise ValueError(f"Tensor must be at least 2D, got shape {shape}")

    rows, cols = shape[-2], shape[-1]
    block_height, block_width = block_structure

    pad_rows = (block_height - rows % block_height) % block_height
    pad_cols = (block_width - cols % block_width) % block_width

    return pad_rows, pad_cols


def maybe_pad_tensor_for_block_quant(
    tensor: torch.Tensor,
    block_structure: tuple[int, int],
) -> torch.Tensor:
    """
    Pad a tensor so its dimensions are divisible by the block size.

    This is essential for FP8 block quantization when dimensions are not
    divisible by block size. The padding ensures that when weights are
    merged in inference frameworks (like vLLM's gate_up_proj), the scale
    tensor blocks align correctly.

    :param tensor: tensor to pad (at least 2D)
    :param block_structure: [block_height, block_width] for block quantization
    :return: padded tensor
    """
    original_shape = tensor.shape

    pad_rows, pad_cols = calculate_block_padding(original_shape, block_structure)

    if pad_rows == 0 and pad_cols == 0:
        return tensor

    # F.pad uses (left, right, top, bottom) for last two dimensions
    padded_tensor = torch.nn.functional.pad(
        tensor, (0, pad_cols, 0, pad_rows), mode="constant", value=0
    )

    return padded_tensor



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
from compressed_tensors.quantization import (
    DEFAULT_QUANTIZATION_FORMAT,
    DEFAULT_QUANTIZATION_METHOD,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
)
from pydantic import ValidationError


def test_basic_config():
    config_groups = {"group_1": QuantizationScheme(targets=[])}
    config = QuantizationConfig(config_groups=config_groups)

    assert config.config_groups == config_groups
    assert config.quant_method == DEFAULT_QUANTIZATION_METHOD
    assert config.format == DEFAULT_QUANTIZATION_FORMAT
    assert config.quantization_status == QuantizationStatus.INITIALIZED
    assert config.global_compression_ratio is None
    assert isinstance(config.ignore, list) and len(config.ignore) == 0


def test_full_config():
    config_groups = {
        "group_1": QuantizationScheme(targets=[]),
        "group_2": QuantizationScheme(targets=[]),
    }
    global_compression_ratio = 3.5
    ignore = ["model.layers.0"]
    quantization_status = "compressed"

    config = QuantizationConfig(
        config_groups=config_groups,
        global_compression_ratio=global_compression_ratio,
        ignore=ignore,
        quantization_status=quantization_status,
    )
    assert config.config_groups == config_groups
    assert config.global_compression_ratio == global_compression_ratio
    assert config.ignore == ignore
    assert config.quantization_status == QuantizationStatus.COMPRESSED


def test_need_config_groups():
    with pytest.raises(ValidationError):
        _ = QuantizationScheme()


@pytest.mark.parametrize(
    "scheme_name",
    ["W8A8", "W8A16", "W4A16", "FP8"],
)
def test_load_scheme_from_preset(scheme_name: str):
    targets = ["Linear"]
    config = QuantizationConfig(config_groups={scheme_name: targets})

    assert scheme_name in config.config_groups
    assert isinstance(config.config_groups[scheme_name], QuantizationScheme)
    assert config.config_groups[scheme_name].targets == targets


def test_to_dict():
    """Test serialization of QuantizationConfig including format"""
    from compressed_tensors.quantization import QuantizationArgs

    config_groups = {
        "group_1": QuantizationScheme(
            targets=["Linear"],
            weights=QuantizationArgs(num_bits=4, symmetric=True, group_size=128),
        ),
        "group_2": QuantizationScheme(
            targets=["Conv2d"],
            weights=QuantizationArgs(num_bits=8),
        ),
    }
    config = QuantizationConfig(
        config_groups=config_groups,
        global_compression_ratio=3.5,
        ignore=["model.layers.0"],
        quantization_status="compressed",
        format="int-quantized",
    )

    # Serialize to dict
    config_dict = config.to_dict()
    assert "config_groups" in config_dict
    assert config_dict["format"] == "int-quantized"
    assert config_dict["quantization_status"] == "compressed"

    # Deserialize from dict
    reloaded = QuantizationConfig.model_validate(config_dict)
    assert config == reloaded



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from itertools import chain
from typing import Iterable, Literal

import torch
import torch.distributed as dist
from compressed_tensors.offload.dist_utils import is_distributed


__all__ = ["get_tensors", "norm_device", "DEFAULT_OFFLOAD_DEVICE"]


DEFAULT_OFFLOAD_DEVICE = torch.device("cpu")


def norm_device(
    device: str | torch.device | None,
) -> torch.device | Literal["disk"] | None:
    """
    Standardize the representation of devices for the purposes of consistency

    - when running with distributed, represent rank device as "cuda"
    - when not running with distributed, "cuda" refers to "cuda:0"
    """
    if device in ("disk", None):
        return device

    device = torch.device(device)

    # (dist) "cuda:R" -> "cuda"
    if is_distributed() and device.index == dist.get_rank():
        device = torch.device(type=device.type, index=None)

    # (non-dist) "cuda" -> "cuda:0"
    if not is_distributed() and device.type == "cuda" and device.index is None:
        device = torch.device(type=device.type, index=0)

    return device


def get_tensors(
    module: torch.nn.Module, recurse: bool = False
) -> Iterable[tuple[str, torch.Tensor | None]]:
    return chain(
        module.named_parameters(recurse=recurse), module.named_buffers(recurse=recurse)
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.compressors.mx_utils import compress_mx_scale
from compressed_tensors.compressors.nvfp4.base import NVFP4PackedCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationType,
)
from compressed_tensors.utils import TensorStateDict


__all__ = ["MXFP4PackedCompressor"]


@BaseCompressor.register(name=CompressionFormat.mxfp4_pack_quantized.value)
class MXFP4PackedCompressor(NVFP4PackedCompressor):
    """
    Compressor for MXFP4 quantized models.

    Overrides scale compression to use log2 encoding (bias-127 exponent).
    Decompression is not implemented for this format.
    """

    @classmethod
    def _compress_scale(
        cls, scale: torch.Tensor, weights: QuantizationArgs
    ) -> torch.Tensor:
        assert weights.scale_dtype is not None
        return compress_mx_scale(scale, weights.scale_dtype)

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict.

        MXFP4 decompression is currently not supported.

        :param state_dict: local-name state dict (weight_packed, weight_scale, …)
        :param scheme: quantization scheme for the weight
        :return: decompressed state dict with weight in float dtype
        """
        raise NotImplementedError("MXFP4 decompression is currently not supported")

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """MXFP4 matches FP4 with group_size=32."""
        return (
            module_type == torch.nn.Linear
            and scheme.weights is not None
            and scheme.weights.num_bits == 4
            and scheme.weights.type == QuantizationType.FLOAT.value
            and scheme.weights.group_size == 32
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import json
import tempfile
from pathlib import Path

import torch
import torch.nn as nn
from compressed_tensors import ModelCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
)
from compressed_tensors.transform import TransformConfig


class DummyLinear(nn.Module):
    """Simple linear module for testing."""

    def __init__(self, in_features=10, out_features=10):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias=False)

    def forward(self, x):
        return self.linear(x)


class TwoLayerModel(nn.Module):
    """Model with two linear layers for testing."""

    def __init__(self):
        super().__init__()
        self.layer1 = nn.Linear(10, 10, bias=False)
        self.layer2 = nn.Linear(10, 10, bias=False)

    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x)
        return x


def create_quantization_config(
    bits=8, type="int", strategy="tensor", format="int-quantized"
):
    """Helper to create a QuantizationConfig for testing."""
    config_dict = {
        "format": format,
        "global_compression_ratio": 1.0,
        "quant_method": "compressed-tensors",
        "config_groups": {
            "group_0": {
                "targets": ["Linear"],
                "weights": {
                    "num_bits": bits,
                    "strategy": strategy,
                    "symmetric": True,
                    "type": type,
                },
            }
        },
    }
    return QuantizationConfig.model_validate(config_dict)


def create_quantization_scheme(
    bits=8, type="int", strategy="tensor", format=None, with_input_activations=False
):
    """Helper to create a QuantizationScheme for testing."""
    input_activations = None
    if with_input_activations:
        input_activations = QuantizationArgs(
            num_bits=bits,
            strategy=strategy,
            symmetric=True,
            type=type,
        )

    return QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            num_bits=bits,
            strategy=strategy,
            symmetric=True,
            type=type,
        ),
        input_activations=input_activations,
        format=format,
    )


class TestModelCompressorCompression:
    """Test compression and decompression functionality."""

    def test_compress_model_basic(self):
        """Test basic model compression."""
        model = DummyLinear()
        # Use pack-quantized format (requires 4 or 8 bits, int type,
        # no input activations)
        scheme = create_quantization_scheme(bits=4, type="int", strategy="channel")
        model.linear.quantization_scheme = scheme
        model.linear.quantization_status = QuantizationStatus.FROZEN
        # Set up quantization parameters
        model.linear.weight_scale = nn.Parameter(
            torch.ones(model.linear.weight.shape[0], 1) * 0.01
        )
        model.linear.weight_zero_point = nn.Parameter(
            torch.zeros(model.linear.weight.shape[0], 1, dtype=torch.int32),
            requires_grad=False,
        )

        q_config = create_quantization_config(bits=4, format="pack-quantized")
        compressor = ModelCompressor(quantization_config=q_config)

        compressor.compress_model(model)

        # Weight should be packed into int32
        assert hasattr(model.linear, "weight_packed")
        assert model.linear.weight_packed.dtype == torch.int32
        assert hasattr(model, "ct_decompress_hook")

    def test_compress_model_skips_non_quantized_modules(self):
        """Test that modules without quantization_scheme are skipped."""
        model = TwoLayerModel()
        # Only quantize layer1
        scheme = create_quantization_scheme(bits=4, type="int", strategy="channel")
        model.layer1.quantization_scheme = scheme
        model.layer1.quantization_status = QuantizationStatus.FROZEN
        model.layer1.weight_scale = nn.Parameter(
            torch.ones(model.layer1.weight.shape[0], 1) * 0.01
        )
        # Use .int() instead of dtype=torch.int32 to avoid Parameter issues
        model.layer1.weight_zero_point = nn.Parameter(
            torch.zeros(model.layer1.weight.shape[0], 1).int(), requires_grad=False
        )

        q_config = create_quantization_config(bits=4, format="pack-quantized")
        compressor = ModelCompressor(quantization_config=q_config)

        layer2_original_dtype = model.layer2.weight.dtype
        compressor.compress_model(model)

        # layer1 should be compressed
        assert hasattr(model.layer1, "weight_packed")
        # layer2 should remain unchanged
        assert model.layer2.weight.dtype == layer2_original_dtype

    def test_compress_decompress_roundtrip(self):
        """Test that compress then decompress recovers original weights."""
        model = DummyLinear()
        scheme = create_quantization_scheme(bits=4, type="int", strategy="channel")
        model.linear.quantization_scheme = scheme
        model.linear.quantization_status = QuantizationStatus.FROZEN
        model.linear.weight_scale = nn.Parameter(
            torch.ones(model.linear.weight.shape[0], 1) * 0.01
        )
        model.linear.weight_zero_point = nn.Parameter(
            torch.zeros(model.linear.weight.shape[0], 1, dtype=torch.int32),
            requires_grad=False,
        )

        q_config = create_quantization_config(bits=4, format="pack-quantized")
        compressor = ModelCompressor(quantization_config=q_config)

        original_weight = model.linear.weight.data.clone()
        compressor.compress_model(model)
        compressor.decompress_model(model)

        # After decompression, weight should be back to float
        assert model.linear.weight.dtype == torch.float32
        # Values should be close (within quantization error for 4-bit)
        diff = torch.abs(original_weight - model.linear.weight.data)
        assert torch.max(diff) < 1.0  # Reasonable threshold for 4-bit quantization

    def test_decompress_hook_triggers_on_forward(self):
        """Test that the decompress hook is triggered on forward pass."""
        model = DummyLinear()
        scheme = create_quantization_scheme(bits=4, type="int", strategy="channel")
        model.linear.quantization_scheme = scheme
        model.linear.quantization_status = QuantizationStatus.FROZEN
        model.linear.weight_scale = nn.Parameter(
            torch.ones(model.linear.weight.shape[0], 1) * 0.01
        )
        model.linear.weight_zero_point = nn.Parameter(
            torch.zeros(model.linear.weight.shape[0], 1, dtype=torch.int32),
            requires_grad=False,
        )

        q_config = create_quantization_config(bits=4, format="pack-quantized")
        compressor = ModelCompressor(quantization_config=q_config)

        compressor.compress_model(model)
        assert hasattr(model.linear, "weight_packed")
        assert hasattr(model, "ct_decompress_hook")

        # Forward pass should trigger decompression
        x = torch.randn(2, 10)
        _ = model(x)

        # After forward, weight should be decompressed
        assert model.linear.weight.dtype == torch.float32
        assert not hasattr(model.linear, "weight_packed")

    def test_compress_model_updates_format_in_config(self):
        """Test that from_pretrained_model infers and sets the format correctly."""
        model = DummyLinear()
        scheme = create_quantization_scheme(bits=4, type="int", strategy="channel")
        model.linear.quantization_scheme = scheme
        model.linear.quantization_status = QuantizationStatus.FROZEN
        model.linear.weight_scale = nn.Parameter(
            torch.ones(model.linear.weight.shape[0], 1) * 0.01
        )
        model.linear.weight_zero_point = nn.Parameter(
            torch.zeros(model.linear.weight.shape[0], 1, dtype=torch.int32),
            requires_grad=False,
        )

        # Use from_pretrained_model which infers the format
        compressor = ModelCompressor.from_pretrained_model(model)

        # Format should be inferred as pack-quantized
        assert compressor.quantization_config.format == CompressionFormat.pack_quantized


class TestModelCompressorConfigUpdate:
    """Test config file update functionality."""

    def test_update_config_creates_file(self):
        """Test that update_config creates config.json if it doesn't exist."""
        with tempfile.TemporaryDirectory() as tmpdir:
            q_config = create_quantization_config()
            compressor = ModelCompressor(quantization_config=q_config)

            compressor.update_config(tmpdir)

            config_path = Path(tmpdir) / "config.json"
            assert config_path.exists()

            with open(config_path, "r") as f:
                config_data = json.load(f)

            assert "quantization_config" in config_data
            assert (
                config_data["quantization_config"]["quant_method"]
                == "compressed-tensors"
            )

    def test_update_config_preserves_existing_data(self):
        """Test that update_config preserves existing config data."""
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "config.json"
            existing_data = {"model_type": "test", "hidden_size": 768}
            with open(config_path, "w") as f:
                json.dump(existing_data, f)

            q_config = create_quantization_config()
            compressor = ModelCompressor(quantization_config=q_config)
            compressor.update_config(tmpdir)

            with open(config_path, "r") as f:
                config_data = json.load(f)

            # Original data should be preserved
            assert config_data["model_type"] == "test"
            assert config_data["hidden_size"] == 768
            # New data should be added
            assert "quantization_config" in config_data

    def test_update_config_with_transform_config(self):
        """Test that update_config includes transform_config."""
        from compressed_tensors.transform import TransformArgs, TransformScheme

        with tempfile.TemporaryDirectory() as tmpdir:
            q_config = create_quantization_config()
            t_config = TransformConfig(
                config_groups={
                    "group_0": TransformScheme(
                        type="hadamard",
                        apply=[
                            TransformArgs(targets=["Linear"], location="weight_input")
                        ],
                    )
                }
            )
            compressor = ModelCompressor(
                quantization_config=q_config, transform_config=t_config
            )

            compressor.update_config(tmpdir)

            config_path = Path(tmpdir) / "config.json"
            with open(config_path, "r") as f:
                config_data = json.load(f)

            assert "quantization_config" in config_data
            assert "transform_config" in config_data["quantization_config"]

    def test_update_config_no_configs(self):
        """Test that update_config does nothing when no configs are present."""
        with tempfile.TemporaryDirectory() as tmpdir:
            compressor = ModelCompressor()
            compressor.update_config(tmpdir)

            config_path = Path(tmpdir) / "config.json"
            # Should not create file if no configs
            assert not config_path.exists()

    def test_update_config_includes_version(self):
        """Test that update_config includes compression version."""
        import compressed_tensors

        with tempfile.TemporaryDirectory() as tmpdir:
            q_config = create_quantization_config()
            compressor = ModelCompressor(quantization_config=q_config)
            compressor.update_config(tmpdir)

            config_path = Path(tmpdir) / "config.json"
            with open(config_path, "r") as f:
                config_data = json.load(f)

            assert "quantization_config" in config_data
            # The version field is added by update_config
            qc = config_data["quantization_config"]
            assert "version" in qc, f"Keys in qc: {qc.keys()}"
            assert qc["version"] == compressed_tensors.__version__


class TestModelCompressorEdgeCases:
    """Test edge cases and error handling."""

    def test_compress_model_infers_format(self):
        """Test that compression infers format when not set on scheme."""
        model = DummyLinear()
        scheme = QuantizationScheme(
            targets=["Linear"],
            weights=QuantizationArgs(
                num_bits=4, type="int", strategy="channel", symmetric=True
            ),
        )
        # No format set - will be inferred
        model.linear.quantization_scheme = scheme
        model.linear.quantization_status = QuantizationStatus.FROZEN
        model.linear.weight_scale = nn.Parameter(
            torch.ones(model.linear.weight.shape[0], 1) * 0.01
        )
        model.linear.weight_zero_point = nn.Parameter(
            torch.zeros(model.linear.weight.shape[0], 1, dtype=torch.int32),
            requires_grad=False,
        )

        q_config = create_quantization_config(bits=4)
        compressor = ModelCompressor(quantization_config=q_config)
        compressor.compress_model(model)

        # Format should be inferred and set
        assert model.linear.quantization_scheme.format is not None
        assert hasattr(model.linear, "weight_packed")

    def test_empty_model(self):
        """Test compression of a model with no quantized modules."""
        model = nn.Sequential()
        q_config = create_quantization_config()
        compressor = ModelCompressor(quantization_config=q_config)

        # Should not raise, just do nothing
        compressor.compress_model(model)
        compressor.decompress_model(model)

    def test_model_with_no_quantization_scheme(self):
        """Test that modules without quantization_scheme are skipped."""
        model = TwoLayerModel()
        # Don't add quantization_scheme to either layer

        q_config = create_quantization_config()
        compressor = ModelCompressor(quantization_config=q_config)

        original_dtype = model.layer1.weight.dtype
        compressor.compress_model(model)

        # Should skip all modules
        assert model.layer1.weight.dtype == original_dtype
        assert model.layer2.weight.dtype == original_dtype



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import torch
import torch.distributed as dist


__all__ = ["is_distributed", "is_rank0"]


def is_rank0() -> bool:
    return not is_distributed() or dist.get_rank() == 0


def is_distributed() -> bool:
    return dist.is_available() and dist.is_initialized()


def init_dist():
    if "TORCHELASTIC_RUN_ID" not in os.environ:
        raise ValueError(
            "Cannot find distributed environment. "
            "Please make sure you are using `torchrun --nproc-per-node ...`."
        )

    rank = int(os.environ["RANK"])
    local_rank = int(os.environ["LOCAL_RANK"])
    world_size = int(os.environ["WORLD_SIZE"])

    device = torch.device(f"cuda:{local_rank}")
    torch.cuda.set_device(device)
    dist.init_process_group(
        backend="nccl",
        init_method="env://",
        rank=rank,
        world_size=world_size,
        device_id=device,
    )
    dist.barrier()


_FP8_DTYPES = (
    torch.float8_e4m3fn,
    torch.float8_e5m2,
    torch.float8_e4m3fnuz,
    torch.float8_e5m2fnuz,
)


def as_broadcastable(tensor: torch.Tensor) -> torch.Tensor:
    """Return a view of the tensor that is compatible with ``dist.broadcast``.

    NCCL does not support broadcasting FP8 dtypes on hardware without sm_90
    (Hopper or later). This function works around the limitation by viewing FP8
    tensors as ``uint8``, which NCCL can broadcast on any hardware. Non-FP8
    tensors are returned unchanged.

    :param tensor: the tensor to prepare for broadcasting
    :return: the original tensor, or a ``uint8`` view if the dtype is FP8
    """
    if tensor.dtype in _FP8_DTYPES:
        return tensor.data.view(torch.uint8)
    else:
        return tensor



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa

from .from_accelerate import from_accelerate
from .to_accelerate import to_accelerate



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import List, Optional

import pytest
import torch
from compressed_tensors.quantization.quant_args import QuantizationArgs
from compressed_tensors.quantization.quant_config import QuantizationStatus
from compressed_tensors.quantization.quant_scheme import QuantizationScheme


@pytest.fixture
def create_quantization_scheme():
    def quantization_scheme(
        targets: List[str],
        weights: Optional[QuantizationArgs] = None,
        input_activations: Optional[QuantizationArgs] = None,
        output_activations: Optional[QuantizationArgs] = None,
    ):
        return QuantizationScheme(
            targets=targets,
            weights=weights,
            input_activations=input_activations,
            output_activations=output_activations,
        )

    return quantization_scheme


@pytest.fixture
def mock_frozen():
    def update_status(model: torch.nn.Module):
        model.quantization_status = QuantizationStatus.FROZEN

    return update_status



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import Optional

from compressed_tensors.config import (
    CompressionFormat,
    SparsityCompressionConfig,
    SparsityStructure,
)


__all__ = ["Sparse24BitMaskConfig"]


@SparsityCompressionConfig.register(name=CompressionFormat.sparse_24_bitmask.value)
class Sparse24BitMaskConfig(SparsityCompressionConfig):
    """
    Configuration for storing a 24 sparse model using
    bytemask compression

    :param global_sparsity: average sparsity of the entire model
    :param sparsity_structure: structure of the sparsity, should always be
        "2:4" for this compression format
    """

    format: str = CompressionFormat.sparse_24_bitmask.value
    global_sparsity: Optional[float] = 0.0
    sparsity_structure: Optional[str] = SparsityStructure.TWO_FOUR.value



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections import OrderedDict

import pytest
import torch
from compressed_tensors.compressors.format import infer_model_format
from compressed_tensors.quantization import preset_name_to_scheme


@pytest.mark.parametrize(
    "preset,expected_format",
    [
        ["W8A8", "int-quantized"],
        ["W8A16", "pack-quantized"],
        ["W4A16", "pack-quantized"],
        ["FP8", "float-quantized"],
    ],
)
def test_infer_quant_format(preset, expected_format):
    quant_scheme = preset_name_to_scheme(preset, targets=["Linear"])

    dummy_model = torch.nn.Sequential(
        OrderedDict(
            [
                ("fc1", torch.nn.Linear(8, 16, bias=True)),
                ("fc2", torch.nn.Linear(16, 32, bias=True)),
                (
                    "block1",
                    torch.nn.Sequential(
                        OrderedDict(
                            [
                                ("fc1", torch.nn.Linear(32, 16, bias=True)),
                                ("fc2", torch.nn.Linear(16, 8, bias=True)),
                            ]
                        )
                    ),
                ),
            ]
        )
    )

    for _, module in dummy_model.named_modules():
        if isinstance(module, torch.nn.Linear):
            module.quantization_scheme = quant_scheme

    assert infer_model_format(dummy_model).value == expected_format



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import gc
from weakref import ref

import pytest
import torch
from compressed_tensors.offload.cache.device import DeviceCache
from tests.test_offload.cache.helpers import (
    _test_delete,
    _test_disable_onloading,
    _test_offload,
    _test_onload,
    _test_onloading,
    _test_shared_attributes,
    _test_tensor_subclass,
)
from tests.test_offload.conftest import assert_device_equal
from tests.testing_utils import requires_gpu


@pytest.fixture()
def onload_device():
    return torch.device("cuda")


@pytest.fixture()
def offload_device():
    return torch.device("cuda")


@pytest.mark.unit
@requires_gpu
def test_delete(offload_device, onload_device, offload_cache):
    _test_delete(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_disable_offloading(onload_device):
    # unlike other device caches, the onload is not garbage collected
    cache = DeviceCache(onload_device)
    cache["weight"] = torch.ones(10)

    outside_onloaded = cache["weight"]
    outside_onloaded_ref = ref(outside_onloaded)
    assert_device_equal(outside_onloaded.device, onload_device)

    with cache.disable_offloading():
        inside_onloaded = cache["weight"]
        inside_onloaded_ref = ref(inside_onloaded)
        assert_device_equal(inside_onloaded.device, onload_device)

        del outside_onloaded
        del inside_onloaded
        gc.collect()

        assert outside_onloaded_ref() is not None  # changed
        assert inside_onloaded_ref() is not None

    assert outside_onloaded_ref() is not None  # changed
    assert inside_onloaded_ref() is not None  # changed


@pytest.mark.unit
@requires_gpu
def test_disable_onloading(offload_device, onload_device, offload_cache):
    _test_disable_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_garbage_collect(onload_device):
    # unlike other device caches, the onload is not garbage collected
    cache = DeviceCache(onload_device)
    cache["weight"] = torch.ones(10)
    onloaded = cache["weight"]

    onloaded_ref = ref(onloaded)
    del onloaded
    gc.collect()
    assert onloaded_ref() is not None  # changed


@pytest.mark.unit
@requires_gpu
def test_offload(offload_device, onload_device, offload_cache):
    _test_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
@requires_gpu
def test_onload(offload_device, onload_device, offload_cache):
    _test_onload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_onloading(offload_device, onload_device, offload_cache):
    _test_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_shared_attributes(offload_device, onload_device, offload_cache):
    _test_shared_attributes(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_tensor_subclass(offload_device, onload_device, offload_cache):
    _test_tensor_subclass(offload_device, onload_device, offload_cache)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import gc
from weakref import ref

import pytest
import torch
import torch.distributed as dist
from compressed_tensors.offload import disable_onloading
from compressed_tensors.offload.cache.dist_device import DistributedDeviceCache
from tests.test_offload.cache.helpers import (
    _test_delete,
    _test_disable_onloading,
    _test_offload,
    _test_onload,
    _test_onloading,
    _test_shared_attributes,
    _test_tensor_subclass,
)
from tests.test_offload.conftest import assert_device_equal, torchrun
from tests.testing_utils import requires_gpu


# Note that tests only require at least 1 gpu
# b/c different ranks can share the same gpu


@pytest.fixture()
def onload_device():
    return torch.device("cuda")


@pytest.fixture()
def offload_device():
    return torch.device("cuda")


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_delete(offload_device, onload_device, offload_cache):
    _test_delete(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_disable_offloading(onload_device):
    # unlike other device caches, the onload is not garbage collected
    cache = DistributedDeviceCache(onload_device)
    cache["weight"] = torch.ones(10)

    outside_onloaded = cache["weight"]
    outside_onloaded_ref = ref(outside_onloaded)
    assert_device_equal(outside_onloaded.device, onload_device)

    with cache.disable_offloading():
        inside_onloaded = cache["weight"]
        inside_onloaded_ref = ref(inside_onloaded)
        assert_device_equal(inside_onloaded.device, onload_device)

        del outside_onloaded
        del inside_onloaded
        gc.collect()

        assert outside_onloaded_ref() is not None  # changed
        assert inside_onloaded_ref() is not None

    assert outside_onloaded_ref() is not None  # changed
    assert inside_onloaded_ref() is not None  # changed


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_disable_onloading(offload_device, onload_device, offload_cache):
    _test_disable_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_garbage_collect(onload_device):
    # unlike other device caches, the onload is not garbage collected
    cache = DistributedDeviceCache(onload_device)
    cache["weight"] = torch.ones(10)
    onloaded = cache["weight"]

    onloaded_ref = ref(onloaded)
    del onloaded
    gc.collect()
    assert onloaded_ref() is not None  # changed


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_offload(offload_device, onload_device, offload_cache):
    _test_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_onload(offload_device, onload_device, offload_cache):
    _test_onload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_onloading(offload_device, onload_device, offload_cache):
    _test_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_shared_attributes(offload_device, onload_device, offload_cache):
    _test_shared_attributes(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_tensor_subclass(offload_device, onload_device, offload_cache):
    _test_tensor_subclass(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_distributed_offload(onload_device):
    cache = DistributedDeviceCache(onload_device)
    tensor = torch.zeros((5, 2))
    cache["tensor"] = tensor

    # check tensor construction
    assert torch.equal(cache["tensor"].cpu(), tensor)
    with disable_onloading():
        assert torch.equal(cache["tensor"].cpu(), tensor)

    # update tensor
    tensor = torch.ones((5, 2))
    cache["tensor"] = tensor

    # check tensor construction
    assert torch.equal(cache["tensor"].cpu(), tensor)
    with disable_onloading():
        assert torch.equal(cache["tensor"].cpu(), tensor)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_distributed_offload_fp8(onload_device):
    """FP8 tensors should broadcast successfully and preserve their dtype"""
    float8_dtypes = [
        torch.float8_e4m3fn,
        torch.float8_e5m2,
        torch.float8_e4m3fnuz,
        torch.float8_e5m2fnuz,
    ]
    for dtype in float8_dtypes:
        cache = DistributedDeviceCache(onload_device)
        tensor = torch.zeros((5, 2), dtype=dtype)
        cache["tensor"] = tensor

        result = cache["tensor"].cpu()
        assert result.shape == tensor.shape
        assert result.dtype == dtype


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_replicated_device_offload(onload_device):
    cache = DistributedDeviceCache(onload_device)
    tensor = torch.empty((5, 2))
    cache["tensor"] = tensor

    # modify the offloaded cpu tensor directly
    tensor = torch.full((5, 2), dist.get_rank())
    cache["tensor"].copy_(tensor)
    dist.barrier()

    # check that the value is affected on all ranks
    assert torch.equal(cache["tensor"].cpu(), tensor)
    with disable_onloading():
        assert torch.equal(cache["tensor"].cpu(), tensor)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import Optional

from compressed_tensors.config import CompressionFormat, SparsityCompressionConfig


__all__ = ["BitmaskConfig"]


@SparsityCompressionConfig.register(name=CompressionFormat.sparse_bitmask.value)
class BitmaskConfig(SparsityCompressionConfig):
    """
    Configuration for storing a sparse model using
    bitmask compression

    :param global_sparsity: average sparsity of the entire model
    :param sparsity_structure: structure of the sparsity, such as
    "unstructured", "2:4", "8:16" etc
    """

    format: str = CompressionFormat.sparse_bitmask.value
    global_sparsity: Optional[float] = 0.0
    sparsity_structure: Optional[str] = "unstructured"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import hashlib
import os
from pathlib import Path

import pytest
import torch
import torch.distributed as dist
from compressed_tensors.offload import (
    disable_onloading,
    from_accelerate,
    is_rank0,
    load_offloaded_model,
)
from compressed_tensors.offload.cache import CPUCache, DeviceCache, DiskCache
from compressed_tensors.offload.convert.from_accelerate import (
    remove_accelerate_from_module,
)
from tests.test_offload.conftest import torchrun
from tests.testing_utils import requires_gpu
from transformers import AutoModelForCausalLM


acclerate = pytest.importorskip("accelerate")


@pytest.mark.unit
@requires_gpu
def test_remove_accelerate_from_module_device(cuda_device):
    # there"s no way to force accelerate to "offload" to cuda. Instead, it just
    # stays on cuda with no hooks
    linear = torch.nn.Linear(5, 5, device="cuda:0")
    assert remove_accelerate_from_module(linear) == (cuda_device, cuda_device, None)
    assert not hasattr(linear, "_hf_hook")

    # test idempotency
    assert remove_accelerate_from_module(linear) == (cuda_device, cuda_device, None)
    assert not hasattr(linear, "_hf_hook")


@pytest.mark.unit
@requires_gpu
def test_remove_accelerate_from_module_cpu(cuda_device):
    from accelerate.big_modeling import dispatch_model

    linear = torch.nn.Linear(5, 5)
    dispatch_model(
        linear,
        {"": "cpu"},
        main_device="cuda",
        state_dict=linear.state_dict(),
        force_hooks=True,
    )
    assert remove_accelerate_from_module(linear) == (
        cuda_device,
        torch.device("cpu"),
        None,
    )
    assert not hasattr(linear, "_hf_hook")


@pytest.mark.unit
@requires_gpu
@pytest.mark.filterwarnings("ignore::UserWarning")
def test_remove_accelerate_from_module_disk(cuda_device, tmp_path):
    # `disk_offload` is a super buggy function, and not reflective of real dispatches
    # `dispatch_model` is also super buggy, and requires at least one cpu device
    from accelerate.big_modeling import dispatch_model

    offload_dir = tmp_path / "offload_dir"
    os.mkdir(offload_dir)

    linear = torch.nn.Linear(5, 5)
    model = torch.nn.Sequential(linear)
    dispatch_model(
        model,
        {"0": "disk", "fake_module": "cpu"},
        main_device="cuda",
        force_hooks=True,
        offload_dir=offload_dir,
    )
    assert remove_accelerate_from_module(linear) == (cuda_device, "disk", offload_dir)
    assert not hasattr(linear, "_hf_hook")


@pytest.mark.unit
@requires_gpu
def test_from_accelerate(cuda_device, tmp_path):
    from accelerate.big_modeling import dispatch_model

    offload_dir = tmp_path / "offload_dir"
    os.mkdir(offload_dir)

    model = torch.nn.Sequential(
        torch.nn.Linear(5, 5), torch.nn.Linear(5, 5), torch.nn.Linear(5, 5)
    )
    if is_rank0():
        dispatch_model(
            model,
            {"0": 0, "1": "cpu", "2": "disk"},
            main_device=str(cuda_device),
            force_hooks=True,
            offload_dir=offload_dir,
        )
    else:
        model.to("meta")

    device_map, _offload_dir = from_accelerate(model)

    # cuda is index agnostic when distributed
    assert device_map == {
        "": (None, None),
        "0": (cuda_device, cuda_device),
        "1": (cuda_device, torch.device("cpu")),
        "2": (cuda_device, "disk"),
    }
    if is_rank0():
        assert _offload_dir == offload_dir
    assert isinstance(model[0]._parameters, DeviceCache)
    assert isinstance(model[1]._parameters, CPUCache)
    assert isinstance(model[2]._parameters, DiskCache)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_from_accelerate_dist(cuda_device, tmp_path):
    test_from_accelerate(cuda_device, tmp_path)


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
@torch.no_grad()
def test_dist_disk_safetensors_update(tmp_path):
    offload_folder = tmp_path / "offload_folder"
    os.makedirs(offload_folder, exist_ok=True)

    with load_offloaded_model():
        model = AutoModelForCausalLM.from_pretrained(
            "Qwen/Qwen3-0.6B",
            dtype="auto",
            device_map="auto_offload",
            max_memory={"cpu": 6e8},
            offload_folder=str(offload_folder),
        )

        # Get the model checkpoint files and hash them
        if dist.get_rank() == 0:
            checkpoint_files = {}
            for file_path in Path(offload_folder).glob("*.safetensors"):
                if not file_path.name.startswith(DiskCache._new_file_prefix):
                    with open(file_path, "rb") as f:
                        checkpoint_files[file_path.name] = hashlib.sha256(
                            f.read()
                        ).hexdigest()

        dist.barrier()

        # Each rank updates a different module's tensor
        rank_0_module = model.model.layers[-1].self_attn.q_proj
        rank_1_module = model.model.layers[-1].self_attn.k_proj
        rank = dist.get_rank()
        if rank == 0:
            rank_0_module.weight *= 0
        elif rank == 1:
            rank_1_module.weight *= 0
            rank_1_module.weight += 1
        dist.barrier()

        # Check that onloaded values are updated across ranks
        assert torch.all(rank_0_module.weight == 0)
        assert torch.all(rank_1_module.weight == 1)

        # Compare model checkpoint files and make sure they're unchanged
        if dist.get_rank() == 0:
            for file_name, original_hash in checkpoint_files.items():
                file_path = offload_folder / file_name
                with open(file_path, "rb") as f:
                    current_hash = hashlib.sha256(f.read()).hexdigest()
                assert current_hash == original_hash

        # Check that the files exist and are not symlinks
        with disable_onloading():
            q_file_path = DiskCache.index[rank_0_module.weight]["safetensors_file"]
            k_file_path = DiskCache.index[rank_1_module.weight]["safetensors_file"]

        if dist.get_rank() == 0:
            assert os.path.exists(q_file_path)
            assert not os.path.islink(q_file_path)
        if dist.get_rank() == 1:
            assert os.path.exists(k_file_path)
            assert not os.path.islink(k_file_path)

        # Delete the parameters
        dist.barrier()
        delattr(rank_0_module, "weight")
        delattr(rank_1_module, "weight")

        # Check that the new files were deleted
        dist.barrier()
        assert not os.path.exists(q_file_path)
        assert not os.path.exists(k_file_path)

        # Compare model checkpoint files again and make sure they're still unchanged
        if dist.get_rank() == 0:
            for file_name, original_hash in checkpoint_files.items():
                file_path = offload_folder / file_name
                with open(file_path, "rb") as f:
                    current_hash = hashlib.sha256(f.read()).hexdigest()
                assert current_hash == original_hash



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections import defaultdict
from typing import Optional

import torch
from compressed_tensors.offload.cache import DiskCache, OffloadCache
from compressed_tensors.offload.convert.helpers import (
    DEFAULT_OFFLOAD_DEVICE,
    get_tensors,
    norm_device,
)
from compressed_tensors.offload.module import remove_module_offload
from compressed_tensors.utils import patch_attr
from loguru import logger


__all__ = ["to_accelerate", "to_accelerate_module"]


def to_accelerate(model: torch.nn.Module) -> dict[str, str]:
    """
    Convert a model from `compressed_tensors` offloading to `accelerate` offloading.
    This is is often called before `PreTrainedModel.save_pretrained`, as without this
    conversion, `save_pretrained` will use excessive memory and device movement.

    :param model: model dispatched with `compressed_tensors` offloading
    :return: accelerate-style device map
    """
    hf_device_map = {}
    hf_disk_index = _to_accelerate_disk_index(model, DiskCache.index)

    for name, module in model.named_modules():
        offload_device_str = to_accelerate_module(module, name, hf_disk_index)
        hf_device_map[name] = offload_device_str

    setattr(model, "hf_device_map", hf_device_map)
    return hf_device_map


def to_accelerate_module(
    module: torch.nn.Module,
    name: Optional[str] = None,
    hf_disk_index: Optional[dict[str, dict[str, str]]] = None,
) -> str:
    """
    Convert a module from `compressed_tensors` offloading to `accelerate` offloading

    :param module: module to convert to accelerate offloading
    :param name: name of module in model
    :param hf_disk_index: accelerate-style disk index to attach to weight loaders
    :return: str of offloaded device. Defaults to cpu if module does not have parameters
    """
    has_accelerate = True
    try:
        from accelerate.hooks import AlignDevicesHook, add_hook_to_module
        from accelerate.utils import OffloadedWeightsLoader, PrefixedDataset
    except ImportError:
        logger.warning(
            "Cannot convert module without `accelerate` installed. This may result "
            "in high memory usage during saving and tied tensors being saved twice",
            log_once=True,
        )
        has_accelerate = False

    offload_device = DEFAULT_OFFLOAD_DEVICE
    if has_accelerate and isinstance(module._parameters, OffloadCache):
        cache = module._parameters
        offload_device = norm_device(cache.offload_device)
        remove_module_offload(module, onload_tensors=False)

        # create weights map
        if isinstance(cache, DiskCache):
            if name is None or hf_disk_index is None:
                raise ValueError(
                    "Must provide `name` and `hf_disk_index` "
                    "to convert disk offloaded module"
                )

            weights_map = PrefixedDataset(
                prefix=f"{name}.",
                dataset=OffloadedWeightsLoader(
                    index=hf_disk_index,
                    save_folder=cache.offload_dir,
                ),
            )
        else:
            weights_map = dict(get_tensors(module, recurse=False))

        # create hook
        hook = AlignDevicesHook(
            execution_device=cache.onload_device,
            offload=True,
            io_same_device=True,
            weights_map=weights_map,
            offload_buffers=True,
            place_submodules=False,
        )

        # add hook (skip onloading)
        with patch_attr(AlignDevicesHook, "init_hook", lambda self, module: module):
            add_hook_to_module(module, hook)

        # skipping init hook => need to populate `original_devices`
        hook.original_devices = {
            name: offload_device if offload_device != "disk" else torch.device("cpu")
            for name, _ in get_tensors(module, recurse=False)
        }

    return str(offload_device)


def _to_accelerate_disk_index(
    model: torch.nn.Module, index: dict[torch.Tensor, dict[str, str]]
) -> dict[str, dict[str, str]]:
    from compressed_tensors.offload import disable_onloading  # circular dependency

    with disable_onloading():
        offloaded_to_key = _invert_dict(model.state_dict(keep_vars=True))

    return {
        key: weight_info
        for offloaded, weight_info in index.items()
        for key in offloaded_to_key[offloaded]
    }


def _invert_dict(d: dict) -> dict:
    inverted = defaultdict(list)
    for key, value in d.items():
        inverted[value].append(key)
    return inverted



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.entrypoints.convert import (
    convert_checkpoint,
    ModelOptNvfp4Converter,
)
from compressed_tensors.quantization import QuantizationArgs, QuantizationType

MODEL_ID = "nvidia/Qwen3-32B-NVFP4"
SAVE_DIR = MODEL_ID.rstrip("/").split("/")[-1]

# Convert modelopt nvfp4 example to compressed-tensors format
convert_checkpoint(
    model_stub=MODEL_ID,
    save_directory=SAVE_DIR,
    converter=ModelOptNvfp4Converter(
        # nvidia/Qwen3-32B-NVFP4's nvfp4-quantized layers, found by inspection
        targets=[
            "re:.*mlp.*\.(gate_up|gate|up|down)_proj$",
            "re:.*self_attn.*\.(q|k|v|o)_proj$",
        ],
        # nvidia/Qwen3-32B-NVFP4's kv_cache_scheme, found by inspection
        kv_cache_scheme=QuantizationArgs(
            num_bits=8, dynamic=False, type=QuantizationType.FLOAT
        ),
    ),
    max_workers=8,
)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.compressors.nvfp4.helpers import (
    pack_fp4_to_uint8,
    unpack_fp4_from_uint8,
)
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationType,
)
from compressed_tensors.quantization.lifecycle.forward import dequantize, quantize
from compressed_tensors.utils import TensorStateDict


__all__ = ["NVFP4PackedCompressor"]


@BaseCompressor.register(name=CompressionFormat.nvfp4_pack_quantized.value)
class NVFP4PackedCompressor(BaseCompressor):
    """
    Compressor for FP4 quantized models.

    Weights of each quantized layer are packed into uint8. Only supports
    symmetric weight compression.
    """

    @classmethod
    def _compress_scale(
        cls, scale: torch.Tensor, weights: QuantizationArgs
    ) -> torch.Tensor:
        assert weights.scale_dtype is not None
        return scale.to(weights.scale_dtype)

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict.

        Quantizes the weight and packs into uint8 as ``weight_packed``.
        Compresses the scale according to ``scheme.weights.scale_dtype``.
        Removes the raw ``weight``.

        :param state_dict: local-name state dict (weight, weight_scale, …)
        :param scheme: quantization scheme for the weight
        :return: compressed state dict
        """
        state_dict = state_dict.copy()
        weight = state_dict.pop("weight")
        scale = state_dict.pop("weight_scale")
        global_scale = state_dict.get("weight_global_scale", None)
        zero_point = state_dict.get("weight_zero_point", None)
        weights = scheme.weights

        quantized_weight = quantize(
            x=weight,
            scale=scale,
            global_scale=global_scale,
            zero_point=zero_point,
            args=weights,
        )
        state_dict["weight_packed"] = pack_fp4_to_uint8(quantized_weight)
        state_dict["weight_scale"] = cls._compress_scale(scale, weights)
        state_dict = cls._remove_symmetric_zp(state_dict, scheme)

        return state_dict

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict.

        Unpacks ``weight_packed`` back to fp4 values and dequantizes.
        Converts ``weight_scale`` back to float for dequantization.

        :param state_dict: local-name state dict (weight_packed, weight_scale, …)
        :param scheme: quantization scheme for the weight
        :return: decompressed state dict with weight in float dtype
        """
        state_dict = state_dict.copy()
        packed = state_dict.pop("weight_packed")
        scale = state_dict.get("weight_scale")
        global_scale = state_dict.get("weight_global_scale", None)

        m, n = packed.shape
        unpacked = unpack_fp4_from_uint8(packed, m, n * 2)

        # Decompress scale back to float
        scale_float = scale.to(unpacked.dtype)

        state_dict["weight"] = dequantize(
            x_q=unpacked,
            scale=scale_float,
            global_scale=global_scale,
            dtype=unpacked.dtype,
        )
        state_dict["weight_scale"] = torch.nn.Parameter(
            scale_float, requires_grad=False
        )

        return state_dict

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """NVFP4 matches FP4 with group_size != 32 (or None)."""
        return (
            module_type == torch.nn.Linear
            and scheme.weights is not None
            and scheme.weights.num_bits == 4
            and scheme.weights.type == QuantizationType.FLOAT.value
            and scheme.weights.group_size == 16
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import pytest
from compressed_tensors.offload import OffloadCache


@pytest.fixture()
def offload_cache(offload_device, onload_device, tmp_path):
    if offload_device == "disk":
        offload_dir = str(tmp_path / "offload_dir")
        os.makedirs(offload_dir)
        return OffloadCache.cls_from_device(offload_device)(
            onload_device, offload_dir=offload_dir
        )
    else:
        return OffloadCache.cls_from_device(offload_device)(onload_device)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
import torch.distributed as dist
from compressed_tensors.offload.cache.device import DeviceCache
from compressed_tensors.offload.dist_utils import as_broadcastable
from compressed_tensors.offload.utils import send_tensors, to_empty


class DistributedDeviceCache(DeviceCache):
    """
    Handles offloading and onloading tensors from/to device memory. Onloading
    tensors is typically a no-op (except when onload device has been modified).

    The device offload is not shared between ranks. When dispatching with this cache,
    the model is replicated across devices.
    """

    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Move a tensor to device, then broadcast data to all other ranks

        :param value: tensor on any device
        :return: tensor on device
        """
        if tensor is None:
            return None

        if dist.get_rank() == 0:
            tensor = super().offload(tensor)

        # materialize meta tensor only if necessary
        elif tensor.device.type == "meta":
            tensor = to_empty(tensor, device=self.offload_device)
        else:
            tensor = send_tensors(tensor, device=self.offload_device)

        dist.broadcast(as_broadcastable(tensor), src=0)
        return tensor



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import gc
from weakref import ref

import torch
from tests.test_offload.conftest import assert_device_equal, assert_tensor_equal


def _test_onloading(offload_device, onload_device, offload_cache):
    tensor = torch.ones(10)
    offload_cache["weight"] = tensor
    onloaded = offload_cache["weight"]

    assert type(onloaded) is type(tensor)
    assert_tensor_equal(onloaded, tensor, onload_device)


def _test_garbage_collect(offload_device, onload_device, offload_cache):
    offload_cache["weight"] = torch.ones(10)
    onloaded = offload_cache["weight"]

    onloaded_ref = ref(onloaded)
    del onloaded
    gc.collect()
    assert onloaded_ref() is None


def _test_offload(offload_device, onload_device, offload_cache):
    tensor = torch.ones(10, device=onload_device)
    offloaded = offload_cache.offload(tensor)
    assert_device_equal(offloaded.device, offload_device)
    assert_tensor_equal(offloaded, tensor, offload_device)


def _test_onload(offload_device, onload_device, offload_cache):
    tensor = torch.ones(10, device=onload_device)
    onloaded = offload_cache.onload(offload_cache.offload(tensor))
    assert_device_equal(onloaded.device, onload_device)
    assert_tensor_equal(onloaded, tensor, onload_device)


def _test_disable_offloading(offload_device, onload_device, offload_cache):
    offload_cache["weight"] = torch.ones(10)

    outside_onloaded = offload_cache["weight"]
    outside_onloaded_ref = ref(outside_onloaded)
    assert_device_equal(outside_onloaded.device, onload_device)

    with offload_cache.disable_offloading():
        inside_onloaded = offload_cache["weight"]
        inside_onloaded_ref = ref(inside_onloaded)
        assert_device_equal(inside_onloaded.device, onload_device)

        del outside_onloaded
        del inside_onloaded
        gc.collect()

        assert outside_onloaded_ref() is None
        assert inside_onloaded_ref() is not None

    assert outside_onloaded_ref() is None
    assert inside_onloaded_ref() is None


def _test_disable_onloading(offload_device, onload_device, offload_cache):
    tensor = torch.ones(10)
    offload_cache.offloaded_values["weight"] = tensor

    with offload_cache.disable_onloading():
        onloaded = offload_cache["weight"]
        assert onloaded is tensor

    assert onloaded is tensor


def _test_delete(offload_device, onload_device, offload_cache):
    offload_cache["weight"] = torch.ones(10)
    onloaded = offload_cache["weight"]
    onloaded_ref = ref(onloaded)

    with offload_cache.disable_offloading():
        del offload_cache["weight"]
        del onloaded
        gc.collect()

        assert onloaded_ref() is None

    assert onloaded_ref() is None


def _test_shared_attributes(offload_device, onload_device, offload_cache):
    assert (
        offload_cache.offloading_disabled is offload_cache.__class__.offloading_disabled
    )
    assert (
        offload_cache.onloading_disabled is offload_cache.__class__.onloading_disabled
    )
    assert (
        offload_cache.keep_onloaded_values
        is offload_cache.__class__.keep_onloaded_values
    )

    assert not hasattr(offload_cache.__class__, "onload_device")
    assert not hasattr(offload_cache.__class__, "offloaded_values")


def _test_tensor_subclass(offload_device, onload_device, offload_cache):
    tensor = torch.ones(10)
    param = torch.nn.Parameter(torch.ones(10), requires_grad=False)
    buffer = torch.nn.Buffer(torch.ones(10))

    offload_cache["tensor"] = tensor
    offload_cache["param"] = param
    offload_cache["buffer"] = buffer

    assert_tensor_equal(offload_cache["tensor"], tensor, onload_device)
    assert_tensor_equal(offload_cache["param"], param, onload_device)
    assert_tensor_equal(offload_cache["buffer"], buffer, onload_device)

    with offload_cache.disable_onloading():
        assert_tensor_equal(offload_cache["tensor"], tensor, offload_device)
        assert_tensor_equal(offload_cache["param"], param, offload_device)
        assert_tensor_equal(offload_cache["buffer"], buffer, offload_device)


def _test_update_offload(offload_device, onload_device, offload_cache):
    # Create initial tensor and offload it
    initial_data = torch.ones(10, device=onload_device)
    offload_cache["weight"] = initial_data

    # Verify initial value
    onloaded = offload_cache["weight"]
    assert_tensor_equal(onloaded, initial_data, onload_device)

    # Update with new data
    new_data = torch.ones(10, device=onload_device) * 2.0
    offload_cache["weight"] = new_data

    # Verify update worked
    updated_onloaded = offload_cache["weight"]
    assert_tensor_equal(updated_onloaded, new_data, onload_device)

    # Verify offloaded tensor was updated in place (not replaced)
    with offload_cache.disable_onloading():
        offloaded = offload_cache["weight"]
        assert_tensor_equal(offloaded, new_data, offload_device)

    # Test update with disable_offloading context
    with offload_cache.disable_offloading():
        offload_cache["weight"] = torch.ones(10, device=onload_device) * 3.0
        cached_onloaded = offload_cache["weight"]
        assert_tensor_equal(cached_onloaded, torch.ones(10) * 3.0, onload_device)

    # Verify update persisted after context exit
    final_onloaded = offload_cache["weight"]
    assert_tensor_equal(final_onloaded, torch.ones(10) * 3.0, onload_device)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import contextlib
from functools import wraps

import torch
from compressed_tensors.offload.cache.base import OffloadCache
from compressed_tensors.offload.utils import send_tensors


def offload_module(
    module: torch.nn.Module,
    onload_device: torch.device | str,
    offload_device: torch.device | str,
    **kwargs,
):
    """
    Offload a module. Any existing parameters or buffers will be offloaded to the
    offload device specified by the `cache`. Accessing module parameters or buffers will
    cause them to be onloaded to the `onload_device`.

    Calling `forward` will result in input tensors being moved to the `onload_device`,
    and any onloaded parameters or buffers will remain onloaded for the duration of
    the forward call if `no_split` is set to `True`.

    :param module: module to offload
    :param onload_device: device used to onload parameters and buffers
    :param offload_device: device used to offload parameters and buffers
    :param \\**kwargs: keyword arguments for cache constructor
    """
    if isinstance(module._parameters, OffloadCache):
        raise ValueError(
            "Attempted to offload a module twice. "
            "Please call `remove_module_offload` first."
        )

    cache_cls = OffloadCache.cls_from_device(offload_device)
    module._parameters = cache_cls.from_mapping(
        module._parameters, onload_device, offload_device, **kwargs
    )
    module._buffers = cache_cls.from_mapping(
        module._buffers, onload_device, offload_device, **kwargs
    )

    original_forward_func = module.forward.__func__
    module._original_forward_func = original_forward_func

    @wraps(original_forward_func)
    def forward(self, *args, **kwargs):
        if not OffloadCache.onloading_disabled and isinstance(
            module._parameters, OffloadCache
        ):
            onload_device = module._parameters.onload_device
            args = send_tensors(args, device=onload_device)
            kwargs = send_tensors(kwargs, device=onload_device)

        return self._original_forward_func(self, *args, **kwargs)

    module.forward = forward.__get__(module)

    return module


def remove_module_offload(module: torch.nn.Module, onload_tensors: bool = False):
    """
    Remove any offloading applied to the module

    :param onload_tensors: Whether to move tensors to the onloaded device, or keep them
        on the offload device. Defaults to False.
    """
    if isinstance(module._parameters, OffloadCache):
        assert isinstance(module._buffers, OffloadCache)

        if onload_tensors:
            module._parameters = {
                name: module._parameters.onload(param)
                for name, param in module._parameters.offloaded_values.items()
            }
            module._buffers = {
                name: module._buffers.onload(param)
                for name, param in module._buffers.offloaded_values.items()
            }
        else:
            module._parameters = module._parameters.offloaded_values
            module._buffers = module._buffers.offloaded_values

        module.forward = module._original_forward_func.__get__(module)
        del module._original_forward_func


@contextlib.contextmanager
def unwrap_offload_forward(module: torch.nn.Module):
    """
    Upon entering, module forward function is unwrapped. Upon exiting the offloading
    wrapper is added again. Any modifications made to the forward function while within
    the context will be reflected upon exiting.
    """
    if hasattr(module, "_original_forward_func"):
        offload_forward = module.forward
        module.forward = module._original_forward_func.__get__(module)
        yield
        module._original_forward_func = module.forward.__func__
        module.forward = offload_forward

    else:
        yield



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Miscelaneous helpers for the quantization lifecycle
"""

from torch.nn import Module


__all__ = [
    "enable_quantization",
    "disable_quantization",
]


def enable_quantization(module: Module):
    module.quantization_enabled = True


def disable_quantization(module: Module):
    module.quantization_enabled = False



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections.abc import Container
from dataclasses import fields, is_dataclass
from itertools import chain
from typing import TypeVar

import torch
from loguru import logger


__all__ = [
    "send_tensors",
    "get_module_device",
    "move_module_tensor",
    "module_size",
    "to_empty",
    "to_tensor",
]

T = TypeVar("T")
TensorCls = TypeVar("TensorCls", bound=torch.Tensor)


def send_tensors(value: T, *args, **kwargs) -> T:
    """
    Recursively identify and move tensors using `torch.Tensor.to`

    :param value: value containing tensors to move
    :param args: arguments to `to`
    :param kwargs: keyword arguments to `to`
    :return: value with moved tensors
    """
    match value:
        case torch.Tensor():
            with torch.no_grad():
                tensor = value.to(*args, **kwargs)

            # special case: avoid changing param pointer when possible
            if tensor.data_ptr() == value.data_ptr():
                return value

            tensor.__class__ = value.__class__
            tensor.__dict__ = value.__dict__.copy()
            return tensor
        case list():
            return [send_tensors(v, *args, **kwargs) for v in value]
        case tuple():
            return tuple(send_tensors(v, *args, **kwargs) for v in value)
        case dict():
            return {k: send_tensors(v, *args, **kwargs) for k, v in value.items()}
        case _ if is_dataclass(value):
            return type(value)(
                **{
                    f.name: send_tensors(getattr(value, f.name), *args, **kwargs)
                    for f in fields(value)
                }
            )
        case _:
            return value


def get_module_device(
    module: torch.nn.Module, default: torch.device | None = None
) -> torch.device:
    """
    Infer the device of a module using the first
    parameter or buffer registered to the module

    :param module: module to check
    :param default: default device if module does not have tensors or buffers
    :return: device of module
    """
    tensor = next(module.parameters(), next(module.buffers(), None))
    if tensor is not None:
        return tensor.device
    elif default is not None:
        return default
    else:
        logger.warning(
            f"Unable to get execution device of {module}, falling back to CPU",
            log_once=True,
        )
        return torch.device("cpu")


def move_module_tensor(
    module: torch.nn.Module,
    name: str,
    device: int | str | torch.device,
):
    """
    Move a module's tensor to a new device

    :param module: module containing tensors to move
    :param name: name of tensor to move
    :param device: new devices
    """
    if name in module._parameters:
        module._parameters[name] = send_tensors(module._parameters[name], device=device)

    elif name in module._buffers:
        module._buffers[name] = send_tensors(module._buffers[name], device=device)


def get_module_sizes(
    model: torch.nn.Module, no_split_modules: Container[str] = tuple()
) -> list[tuple[torch.nn.Module, int]]:
    """
    Returns a list of modules and their sizes. Only non-splittable modules are returned.
    Non-splittable modules are modules specified by `no_split_modules` or modules with
    direct parameters.

    :param model: model to get sizes from
    :param no_split_modules: module class names which cannot be split
    :return: list of modules and their sizes
    """
    module_sizes = []

    def dfs(module: torch.nn.Module):
        # modules with direct parameters cannot be split
        # otherwise, submodules could return a device that is different than params
        direct_size = module_size(module, recurse=False)
        no_split = module.__class__.__name__ in no_split_modules or direct_size > 0

        total_size = module_size(module, recurse=no_split)
        if total_size > 0:
            module_sizes.append((module, total_size))

        if not no_split:
            for child in module.children():
                dfs(child)

    dfs(model)

    return module_sizes


def module_size(module: torch.nn.Module, recurse: bool = True) -> int:
    """
    Get the size of the module's parameters and buffers in bytes

    :param module: module to check
    :param recurse: whether calculate recursive size, or only direct tensors
    :return: total size of module parameters and buffers
    """
    from compressed_tensors.offload import disable_onloading

    with disable_onloading():
        tensors = chain(
            module.parameters(recurse=recurse), module.buffers(recurse=recurse)
        )
        return sum((tensor.nbytes for tensor in tensors), 0)


def to_empty(tensor: TensorCls, **kwargs) -> TensorCls:
    """
    Create an empty tensor like the given tensor, with the same tensor subclass
    and dict values

    :param tensor: tensor to create empty like
    :return: empty tensor
    """
    empty = torch.empty_like(tensor.data, **kwargs)
    empty.__class__ = tensor.__class__
    empty.__dict__ = tensor.__dict__.copy()
    return empty


def to_tensor(dst: torch.Tensor, src: TensorCls) -> TensorCls:
    """
    Copy the subclass, dict, and attributes into `dst` from `src`

    :param dst: tensor to copy attributes into
    :param src: tensor to copy attributes from
    :return: dst tensor with copied attributes
    """
    dst.__class__ = src.__class__
    dst.__dict__ = src.__dict__.copy()
    dst.requires_grad = src.requires_grad
    return dst



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors.offload import offload_module
from compressed_tensors.quantization import (
    QuantizationMetadata,
    initialize_module_for_quantization,
    preset_name_to_scheme,
)


@pytest.mark.parametrize("offloaded", (True, False))
def test_clear(offloaded):
    module = torch.nn.Linear(16, 16)
    scheme = preset_name_to_scheme("NVFP4", ["Linear"])
    base_forward = module.forward

    # offload module
    if offloaded:
        offload_module(module, "cpu", "cpu")
        offloaded_forward = module.forward
        assert module._original_forward_func is base_forward.__func__

    # add quantized forward (inside offloaded forward)
    initialize_module_for_quantization(module, scheme)
    qparams = ["weight_scale", "weight_global_scale", "input_global_scale"]

    for name in qparams:
        assert hasattr(module, name)

    if offloaded:
        assert module.forward is offloaded_forward
        assert module._original_forward_func.__wrapped__ is base_forward.__func__
    else:
        assert module.forward.__wrapped__ is base_forward.__func__

    # remove quantized forward
    QuantizationMetadata.clear_quantization(module)

    for name in qparams:
        assert not hasattr(module, name)

    if offloaded:
        assert module.forward is offloaded_forward
        assert module._original_forward_func is base_forward.__func__
    else:
        assert module.forward.__func__ is base_forward.__func__



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from tests.test_offload.cache.helpers import (
    _test_delete,
    _test_disable_offloading,
    _test_disable_onloading,
    _test_garbage_collect,
    _test_offload,
    _test_onload,
    _test_onloading,
    _test_shared_attributes,
    _test_tensor_subclass,
)
from tests.testing_utils import requires_gpu


@pytest.fixture()
def onload_device():
    return torch.device("cuda")


@pytest.fixture()
def offload_device():
    return torch.device("cpu")


@pytest.mark.unit
@requires_gpu
def test_delete(offload_device, onload_device, offload_cache):
    _test_delete(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_disable_offloading(offload_device, onload_device, offload_cache):
    _test_disable_offloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_disable_onloading(offload_device, onload_device, offload_cache):
    _test_disable_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_garbage_collect(offload_device, onload_device, offload_cache):
    _test_garbage_collect(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_offload(offload_device, onload_device, offload_cache):
    _test_offload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
@requires_gpu
def test_onload(offload_device, onload_device, offload_cache):
    _test_onload(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_onloading(offload_device, onload_device, offload_cache):
    _test_onloading(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_shared_attributes(offload_device, onload_device, offload_cache):
    _test_shared_attributes(offload_device, onload_device, offload_cache)


@pytest.mark.unit
@requires_gpu
def test_tensor_subclass(offload_device, onload_device, offload_cache):
    _test_tensor_subclass(offload_device, onload_device, offload_cache)
